/*
 * Matrix transformation handler.
 *
 * This file handles the composition of new matrices from the 
 * World, View and Projection matrices.
 */

#ifndef __RC_MATRIXTRANSFORMATION_H__
#define __RC_MATRIXTRANSFORMATION_H__

/*------------------------------------------------------------[ Defines ]----*/

#define N_TRANSFORMATION_MATRICES			24

#define N_FRAME_CONSTANT_MATRICES			12

static const char *transformationMatrixNames[N_TRANSFORMATION_MATRICES] = {
	"View",					"ViewInverse",					"ViewTranspose",					"ViewInverseTranspose",
	"Projection",			"ProjectionInverse",			"ProjectionTranspose",				"ProjectionInverseTranspose",
	"ViewProjection",		"ViewProjectionInverse",		"ViewProjectionTranspose",			"ViewProjectionInverseTranspose",
	"World",				"WorldInverse",					"WorldTranspose",					"WorldInverseTranspose",
	"WorldView",			"WorldViewInverse",				"WorldViewTranspose",				"WorldViewInverseTranspose",
	"WorldViewProjection",	"WorldViewProjectionInverse",	"WorldViewProjectionTranspose",		"WorldViewProjectionInverseTranspose"
};

static const char *normalMatrixName = "NormalMatrix";

#define MAT_VIEW				0x00
#define MAT_PROJ				0x04
#define MAT_VIEW_PROJ			0x08
#define MAT_WORLD				0x0C
#define MAT_WORLD_VIEW			0x10
#define MAT_WORLD_VIEW_PROJ		0x14

#define MAT_BASE				0x00
#define MAT_INV					0x01
#define MAT_T					0x02
#define MAT_INV_T				0x03

/* Use this to optimize projection matrix calculations */
typedef enum {

	PROJECTION_TYPE_FREE_TRANSFORM = 0,
	PROJECTION_TYPE_PERSPECTIVE,
	PROJECTION_TYPE_ORTHOGONAL,	

	PROJECTION_TYPE_LAST

} ProjectionType_t;

/*---------------------------------------------------------------------------*/

class MatrixTransformation {
	friend class SceneGraph;
	friend class World;
	friend class Transformable;
	friend class Geometry;
	friend class TriangleMesh;

/*----------------------------------------------------------[ Functions ]----*/

	public:
		MatrixTransformation();
		~MatrixTransformation();

		/* Set matrices prior to a new frame rendering */
		void prepareFrame(
				mat4f				*nview,
				mat4f				*nprojection,
				mat4f				*nview_projection,
				vec3f				view_position,
				ProjectionType_t	projection_type
			);

		/*
		 * Call after each drawSelf()-call in nodes. This derefs the matrices
		 * that are multiplied by World, and are thus invalid for the next
		 * node to be rendered.
		 * Note: It's unnecessary to call this for nodes that are not
		 * Transformables.
		 */
		void finishNode();

		/* Assemble the matrix with index from transformationMatrixNames */
		mat4f *assembleMatrix(
				const u32			index
			);

		/* 
		 * Find, assemble and push matrices to shader, if the shader wants
		 * them.
		 */
		void setTransformationMatrices(
				ShaderProgram		*shaderProgram
			);

/*------------------------------------------------------------[ Members ]----*/

	private:

		union {
			struct {
				/* View matrix */
				mat4f		*view;
				mat4f		*viewInverse;
				mat4f		*viewTranspose;
				mat4f		*viewInverseTranspose;

				/* Projection matrix */
				mat4f		*projection;
				mat4f		*projectionInverse;
				mat4f		*projectionTranspose;
				mat4f		*projectionInverseTranspose;

				/* ViewProjection matrix */
				mat4f		*viewProjection;
				mat4f		*viewProjectionInverse;
				mat4f		*viewProjectionTranspose;
				mat4f		*viewProjectionInverseTranspose;

				/* World matrix */
				mat4f		*world;
				mat4f		*worldInverse;
				mat4f		*worldTranspose;
				mat4f		*worldInverseTranspose;

				/* WorldView matrix */
				mat4f		*worldView;
				mat4f		*worldViewInverse;
				mat4f		*worldViewTranspose;
				mat4f		*worldViewInverseTranspose;

				/* WorldViewProjection matrix */
				mat4f		*worldViewProjection;
				mat4f		*worldViewProjectionInverse;
				mat4f		*worldViewProjectionTranspose;
				mat4f		*worldViewProjectionInverseTranspose;
			};

			mat4f			*transformationMatrices[N_TRANSFORMATION_MATRICES];
		};

		/*
		 * Matrix pool, so that we don't need to constantly allocate and
		 * deallocate a bunch of memory.
		 */
		mat4f			matListFrame[N_TRANSFORMATION_MATRICES >> 1];
		u32				nextFreeFrame;

		mat4f			matListNode[N_TRANSFORMATION_MATRICES >> 1];
		u32				nextFreeNode;

		ProjectionType_t projectionType;

		/* Normal matrix */
		mat3f			normal;

		/* View info */
		vec3f			viewPosition;
};

#endif /* __RC_MATRIXTRANSFORMATION_H__ */

